/*
 * Copyright (C) 2004 Claudio "KLaN" Ciccani <klan82@cheapnet.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#include <pthread.h>

#include <directfb.h>

#include <xine.h>
#include <xine/xineutils.h>


#ifndef DF_XINE_H
#define DF_XINE_H


#define THIS  "df_xine"



typedef struct
{
     int                          lid;

     IDirectFBDisplayLayer       *layer;
     DFBDisplayLayerCapabilities  caps;
     DFBDisplayLayerConfig        config;      /* current configuration */
     DFBDisplayLayerBufferMode    buffermode;  /* selected buffermode */

     IDirectFBSurface            *surface;
     IDirectFBSurface            *buffer;
     DFBRectangle                 drect;
     DFBSurfaceFlipFlags          flipflags;   /* DSFLIP_WAITFORSYNC or none */

     struct
     {
          int                     width;
          int                     height;
          double                  ratio;
     } screen;

     struct
     {
          int                     width;
          int                     height;
          double                  ratio;
     } video;

     char                         update;

     int                          gmode;       /* Visualization/Video */

} DFXVideoContext;


typedef struct
{
     pthread_t                    thread;
     pthread_mutex_t              mutex;
     pthread_cond_t               cond;
     xine_event_queue_t          *queue;
     
     int                          magic;
     struct timeval               time;
     
     int                          pos;
     int                          len;
     int                          status;
     int                          speed;
     
     int                          volume;
     int                          brightness;
     int                          contrast;
     int                          saturation;
     int                          hue;
     double                       ratio;
     int                          ratio_id;

} DFXVideoControl;


typedef struct
{
     int                          dst_width;
     int                          dst_height;

     int                          fontsize;

     struct
     {
          char                    active;
          xine_osd_t             *img;
          xine_osd_t             *text;
          int                     w;
          int                     h;
     } bar;

     struct
     {
          char                    active;
          xine_osd_t             *text;
          int                     w;
          int                     h;
     } info;

} DFXDisplay;


typedef struct _DFXMrl DFXMrl;

struct _DFXMrl
{
     char                        *mrl;

     DFXMrl                      *next;
     DFXMrl                      *prev;
};


typedef struct
{
     DFXMrl                      *first;
     DFXMrl                      *last;
     DFXMrl                      *cur;
     int                          repeat;
     int                          repeated;

} DFXMedia;


typedef struct
{
     int                          verbosity;
     int                          stdctl;	
     char                        *cfg;

     IDirectFB                   *dfb;
     IDirectFBEventBuffer        *buffer;
     IDirectFBDisplayLayer       *primary;
     IDirectFBSurface            *bg;

     xine_t                      *xine;
     xine_video_port_t           *vo;
     xine_audio_port_t           *ao;
     xine_stream_t               *stream;

     xine_post_t                 *post;
     const char                  *post_plugin;

     DFXVideoContext              ctx;

     DFXVideoControl              ctl;

     DFXDisplay                   dpy;

     DFXMedia                     media;

     int                          quit;

} DFXCore;



typedef struct
{
     DFBInputDeviceKeySymbol      key;    /* modifier */
     DFBInputDeviceKeySymbol      subkey;

     short                        event;
     short                        actid;
     short                        param;

     const char                  *name[2];
     const char                  *help;
     const char                  *stdctl;

} DFXKeyControl;
     



typedef struct
{
     IDirectFBSurface *surface;

     void            (*output_cb) ( void *data, int width, int height,
                         double ratio, DFBRectangle *dest_rect );
     void             *cdata;

} dfb_visual_t;


typedef struct
{
     DVFrameCallback  frame_cb;
     void            *cdata;

} dfb_frame_callback_t;



#define DGM_VIDEO          1
#define DGM_ANIM           2

#define DAN_ACQUIRE        1
#define DAN_RELEASE        2

#define DAI_QUIT           1
#define DAI_TOGGLE_PLAY    2
#define DAI_PLAY           3
#define DAI_TOGGLE_PAUSE   4
#define DAI_SEEK           5
#define DAI_SPEED          6
#define DAI_VOLUME         7
#define DAI_BRIGHTNESS     8
#define DAI_CONTRAST       9
#define DAI_HUE           10
#define DAI_SATURATION    11
#define DAI_RATIO         12
#define DAI_AUDIO_CHANNEL 13
#define DAI_SPU_CHANNEL   14
#define DAI_OSD_HIDE      15

#define DAP_RESET          0
#define DAP_MINUS         -1
#define DAP_PLUS          +1
#define DAP_RATIO_SQUARE   1
#define DAP_RATIO_43       2
#define DAP_RATIO_169      3
#define DAP_RATIO_DVB      4
#define DAP_SEEK_0         0
#define DAP_SEEK_1        10
#define DAP_SEEK_2        20
#define DAP_SEEK_3        30
#define DAP_SEEK_4        40
#define DAP_SEEK_5        50
#define DAP_SEEK_6        60
#define DAP_SEEK_7        70
#define DAP_SEEK_8        80
#define DAP_SEEK_9        90

#define DDF_BAR            1
#define DDF_INFO           2
#define DDF_BOTH           3



#define DFX_MAGIC( event, notice, action, param ) \
                   ((event << 18) | (notice << 16) | (action << 8) | (param & 0xff))

#define DFX_EVENT( magic )   ((magic >> 18) & 0x3fff)
     
#define DFX_NOTICE( magic )  ((magic >> 16) & 0x0003)

#define DFX_ACTION( magic )  ((magic >>  8) & 0x00ff)

#define DFX_PARAM( magic )   ((char) (magic & 0x00ff))




#define DFBCHECK( x... ) \
{\
     int err = x;\
     if (err != DFB_OK)\
     {\
          fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ );\
          DirectFBError( #x, err );\
          dfx_exit( EXIT_FAILURE );\
     }\
}

#define SAY( fmt, ... ) \
{\
     if (this->verbosity)\
          fprintf( stderr, THIS ": " fmt, ## __VA_ARGS__ );\
}

#define ONCE( fmt, ... ) \
{\
     static int once = 1;\
     if (this->verbosity && once)\
     {\
          fprintf( stderr, THIS ": " fmt, ## __VA_ARGS__ );\
          once = 0;\
     }\
}

#define DBUG( fmt, ... ) \
{\
     if (this->verbosity >= 2 )\
          fprintf( stderr, THIS ": " fmt, ## __VA_ARGS__ );\
}

#define FATAL( fmt, ... ) \
{\
     fprintf( stderr, THIS "[!!]: " fmt, ## __VA_ARGS__ );\
     dfx_exit( EXIT_FAILURE );\
}

#define DX_ALLOC( size ) \
({\
     register void *__mem;\
     __mem = calloc( 1, (size) );\
     if (!__mem)\
          FATAL( "calloc( 1, %i ) failed at line %i of %s\n",\
               (size), __LINE__, __FILE__ );\
     __mem;\
})

#define DX_STRDUP( str ) \
({\
     register void *__str;\
     __str = strdup( str );\
     if (!__str)\
          FATAL( "strdup( \"%s\" ) failed at line %i of %s\n",\
               str, __LINE__, __FILE__ );\
     __str;\
})

#define DX_TOSTRING( fmt, ... ) \
({\
     char *__str = NULL;\
     if (asprintf( &__str, fmt, ## __VA_ARGS__ ) < 1 || !__str)\
          FATAL( "asprintf( %p, \"%s\", %s ) failed at line %i of %s\n",\
               &__str, fmt, # __VA_ARGS__, __LINE__, __FILE__ );\
     __str;\
})





void
dfx_playlist_append( DFXCore *this, char *mrl );

void
dfx_playlist_discard( DFXCore *this );

void
dfx_playlist_skip( DFXCore *this, int rel );

void
dfx_playlist_free( DFXCore *this );


void
dfx_context_create( DFXCore *this );

void
dfx_context_update( DFXCore *this, int width, int height );

void
dfx_context_set_gmode( DFXCore *this, int mode );

void
dfx_context_release( DFXCore *this );


void
dfx_control_start( DFXCore *this );

void
dfx_control_action( DFXCore *this, int magic );

void
dfx_control_release( DFXCore *this );


void
dfx_display_bar( DFXCore *this, char *text, int pos );

void
dfx_display_text( DFXCore *this, const char *text, ... );

void
dfx_display_status( DFXCore *this, int update );

void
dfx_display_hide( DFXCore *this, int force );

void
dfx_display_release( DFXCore *this );


void
dfx_exit( int status );




#endif /* DF_XINE_H */

