/*
 * Copyright (C) 2004 Claudio "KLaN" Ciccani <klan82@cheapnet.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <directfb.h>

#include <direct/mem.h>

#include "dfblogo.h"


#define THIS  "sample1"

#define DFBCHECK( x... ) \
{\
	int err = x;\
	if (err != DFB_OK)\
	{\
		fprintf( stderr, "%s <%d>:\n\t", __FILE__, __LINE__ );\
		DirectFBErrorFatal( #x, err );\
		goto EXIT;\
	}\
}

#define ADDS( x, s )  \
{\
	int __n = x + s;\
	__n = (__n < 0) ? 0 : ((__n > 0xffff) ? 0xffff : __n);\
	x = __n;\
}


struct _windata
{
	IDirectFBWindow   *window;
	IDirectFBSurface  *surface;
	int                id;
};

struct _mydata
{
	IDirectFBSurface  *logo;
	IDirectFBSurface  *frame;
	struct _windata   *wins;
	int                nwins;
};


static const struct
{
	const char            *name;
	DFBSurfacePixelFormat  format;

} pixels[] =
{
	{"YUY2",     DSPF_YUY2},
	{"UYVY",     DSPF_UYVY},
	{"YV12",     DSPF_YV12},
	{"I420",     DSPF_I420},
	{"RGB332",   DSPF_RGB332},
	{"ARGB1555", DSPF_ARGB1555},
	{"RGB16",    DSPF_RGB16},
	{"RGB24",    DSPF_RGB24},
	{"RGB32",    DSPF_RGB32},
	{"ARGB",     DSPF_ARGB},
	{"AiRGB",    DSPF_AiRGB}
};






static int
frame_cb( void *cdata )
{
	struct _mydata  *data    = (struct _mydata*) cdata;
	struct _windata *wins    = data->wins;
	static DFBColor  color   = {0xcc, 0xcc, 0xcc, 0xff};
	int              sw, sh;
	int              dw, dh;
	int              i;
		
	data->frame->GetSize( data->frame, &sw, &sh );
	
	for (i = 0; i < data->nwins; i++)
	{
		wins[i].window->GetSize( wins[i].window, &dw, &dh );

		if (dw != sw || dh != sh)
			wins[i].surface->StretchBlit( wins[i].surface,
						data->frame, NULL, NULL );
		else
			wins[i].surface->Blit( wins[i].surface,
						data->frame, NULL, 0, 0 );

		if (data->logo)
		{
			wins[i].surface->SetColor( wins[i].surface, 
					color.r, color.g, color.b, 0xff );
			wins[i].surface->SetBlittingFlags( wins[i].surface,
					DSBLIT_BLEND_ALPHACHANNEL | DSBLIT_COLORIZE );
			wins[i].surface->Blit( wins[i].surface, data->logo, NULL,
					7, (dh - dfblogo_desc.height - 7) );
			wins[i].surface->SetBlittingFlags( wins[i].surface, DSBLIT_NOFX );
		}

		wins[i].surface->Flip( wins[i].surface, NULL, 0 );
	}

	color.r -= 2;
	color.g += 1;
	color.b -= 2;

	return 0;
}


static void
seek( IDirectFBVideoProvider       *video,
      DFBVideoProviderCapabilities  caps,
      double                        step )
{
	static double length = 0.0;
	double        pos    = 0.0;

	if (!(caps & DVCAPS_SEEK))
		return;

	if (!length)
		video->GetLength( video, &length );

	video->GetPos( video, &pos );

	pos += step;
	if (pos < 0.0)
		pos = 0.0;

	video->SeekTo( video, pos );
}


static void
adjust_color( IDirectFBVideoProvider       *video,
              DFBVideoProviderCapabilities  caps,
              DFBColorAdjustmentFlags       flags,
              int                           step )
{
	DFBColorAdjustment adj = { .flags = 0 };

	if (video->GetColorAdjustment( video, &adj ) != DFB_OK)
		return;

	if ((flags & DCAF_BRIGHTNESS) && (caps & DVCAPS_BRIGHTNESS))
	{
		adj.flags |= DCAF_BRIGHTNESS;
		ADDS( adj.brightness, step );
	}

	if ((flags & DCAF_CONTRAST) && (caps & DVCAPS_CONTRAST))
	{
		adj.flags |= DCAF_CONTRAST;
		ADDS( adj.contrast, step );
	}

	if ((flags & DCAF_HUE) && (caps & DVCAPS_HUE))
	{
		adj.flags |= DCAF_HUE;
		ADDS( adj.hue, step );
	}

	if ((flags & DCAF_SATURATION) && (caps & DVCAPS_SATURATION))
	{
		adj.flags |= DCAF_SATURATION;
		ADDS( adj.saturation, step );
	}

	if (adj.flags)
		video->SetColorAdjustment( video, &adj );
}


static void
usage( void )
{
	puts( "\nUsage: " THIS " [options] video-file" );
	puts( "       -h,--help               show this help" );
	puts( "       --format=<pixelformat>  set the pixelformat to use" );
	puts( "                               (one between YUY2, UYVY, YV12," );
	puts( "                                         I420, RGB332, ARGB1555," );
	puts( "                                         RGB16, RGB24, RGB32," );
	puts( "                                         ARGB, AiRGB)" );
	puts( "       --windows=N             play video on N windows (default:1, maximum:10)" );
	puts( "       --no-logo               do not display DirectFB logo in the" );
	puts( "                               lower-left corner of the window\n" );
	puts( "       Use:" );
	puts( "           ESC,Q,q   to quit" );
	puts( "           P,p       to toggle pause" );
	puts( "           ->,<-     to seek" );
	puts( "           B,b + ->  to increase brightness" );
	puts( "           B,b + <-  to decrease brightness" );
	puts( "           C,c + ->  to increase contrast" );
	puts( "           C,c + <-  to decrease contrast" );
	puts( "           S,s + ->  to increase saturation" );
	puts( "           S,s + <-  to decrease saturation" );
	puts( "           H,h + ->  to increase hue" );
	puts( "           H,h + <-  to decrease hue\n" );
	exit( EXIT_FAILURE );
}


int
main( int argc, char** argv )
{
	IDirectFB                    *dfb       = NULL;
	IDirectFBEventBuffer         *buffer    = NULL;
	IDirectFBVideoProvider       *video     = NULL;
	DFBVideoProviderCapabilities  caps      = 0;
	IDirectFBSurface             *frame     = NULL;
	IDirectFBSurface             *logo      = NULL;
	IDirectFBDisplayLayer        *layer     = NULL;
	struct _windata              *wins      = NULL;
	DFBSurfacePixelFormat         format    = 0;
	static struct _mydata         data;
	const char                   *mrl       = NULL;
	int                           disp_logo = 1;
	int                           nwins     = 1;
	int                           i;

	if (argc < 2)
		usage();

	DirectFBInit( &argc, &argv );
	
	for (i = 1; i < argc; i++)
	{
		char *option = argv[i];

		if (!option || !*option)
			continue;

		if (*option == '-')
		{
			option++;

			if (!strcmp( option, "h" ) || !strcmp( option, "-help" ))
				usage();

			if (!strncmp( option, "-format=", strlen( "-format=" ) ))
			{
				int j;

				option += strlen( "-format=" );

				for (j = 0; j < (sizeof( pixels ) /
						sizeof( pixels[0] )); j++)
				{
					if (!strcmp( option, pixels[j].name ))
					{
						format = pixels[j].format;
						fprintf( stderr, THIS 
							": using pixelformat %s\n", 
							pixels[j].name );
						break;
					}
				}

				if (!format)
				{
					fprintf( stderr, THIS 
						": unsupported pixelformat '%s'\n", 
						option );
					exit( EXIT_FAILURE );
				}
			} else
			if (!strncmp( option, "-windows=", strlen( "-windows=" ) ))
			{
				option += strlen( "-windows=" );
				if (isdigit( *option ))
					nwins = strtol( option, NULL, 10 );
				nwins = (nwins < 1) ? 1 : ((nwins > 10) ? 10 : nwins );
			}
			if (!strncmp( option, "-no-logo", strlen( "-no-logo" ) ))
			{
				disp_logo = 0;
			}
			
		} else
		if (i == (argc - 1))
		{
			mrl = option;
		}
	}

	if (!mrl || !*mrl)
		usage();


	DFBCHECK( DirectFBCreate( &dfb ) );

	DFBCHECK( dfb->GetDisplayLayer( dfb, DLID_PRIMARY, &layer ) );

	DFBCHECK( dfb->CreateVideoProvider( dfb, mrl, &video ) );
	DFBCHECK( video->GetCapabilities( video, &caps ) );

	if (disp_logo)
		DFBCHECK( dfb->CreateSurface( dfb, &dfblogo_desc, &logo ) );

	dfb->CreateEventBuffer( dfb, &buffer );

	{
		DFBAccelerationMask   mask;
		DFBSurfaceDescription sdesc;
		DFBWindowDescription  wdesc;
		
		video->GetSurfaceDescription( video, &sdesc );

		sdesc.flags = DSDESC_CAPS | DSDESC_WIDTH | DSDESC_HEIGHT;
		sdesc.caps  = DSCAPS_SYSTEMONLY;

		wdesc.flags  = DWDESC_POSX | DWDESC_POSY |
				DWDESC_WIDTH | DWDESC_HEIGHT;
		wdesc.posx   = 0;
		wdesc.posy   = 0;
		wdesc.width  = sdesc.width;
		wdesc.height = sdesc.height;

		if(format)
		{
			sdesc.flags       |= DSDESC_PIXELFORMAT;
			wdesc.flags       |= DWDESC_PIXELFORMAT;
			sdesc.pixelformat  = format;
			wdesc.pixelformat  = format;
		}

		wins = (struct _windata*) D_CALLOC( 1, nwins * sizeof( struct _windata ) );
		if (!wins)
			goto EXIT;

		for (i = 0; i < nwins; i++, wdesc.posx += 10, wdesc.posy += 5)
		{
			DFBCHECK( layer->CreateWindow( layer, &wdesc, &wins[i].window ) );
			wins[i].window->GetID( wins[i].window, &wins[i].id );
			wins[i].window->SetOpacity( wins[i].window, 0xff );
			wins[i].window->AttachEventBuffer( wins[i].window, buffer );
			
			DFBCHECK( wins[i].window->GetSurface( wins[i].window, &wins[i].surface ) );
			wins[i].surface->Clear( wins[i].surface, 0, 0, 0, 0xff );
			wins[i].surface->Flip( wins[i].surface, NULL, 0 );
		}

		/* test acceleration on first surface */
		wins[0].surface->GetAccelerationMask( wins[0].surface,
						      wins[0].surface, &mask );
		if (mask & DFXL_STRETCHBLIT)
			sdesc.caps = DSCAPS_VIDEOONLY;

		DFBCHECK( dfb->CreateSurface( dfb, &sdesc, &frame ) );
	}

	data.logo  = logo;
	data.frame = frame;
	data.wins  = wins;
	data.nwins = nwins;

	DFBCHECK( video->PlayTo( video, frame, NULL, frame_cb, (void*) &data ) );

	while (1)
	{
		static int              is_playing = 1;
		static int              adj        = 0;
		static IDirectFBWindow *active     = NULL;
		static int              moved      = 0;
		static int              startx;
		static int              starty;
		static int              endx;
		static int              endy;
		DFBWindowEvent          evt;

		buffer->WaitForEvent( buffer );

		while (buffer->GetEvent( buffer, DFB_EVENT( &evt ) ) == DFB_OK)
		{
			switch (evt.type)
			{
				case DWET_KEYDOWN:
				{
					switch (tolower( evt.key_symbol ))
					{
						case DIKS_ESCAPE:
						case DIKS_SMALL_Q:
							goto EXIT;
						break;

						case DIKS_SMALL_P:
						{
							if(is_playing)
							{
								video->Stop( video );
								is_playing = 0;
								break;
							}
							
							DFBCHECK( video->PlayTo( video, frame, 
									NULL, frame_cb,	(void*) &data ) );
							is_playing = 1;
						}
						break;

						case DIKS_SMALL_B:
							adj |= DCAF_BRIGHTNESS;
						break;

						case DIKS_SMALL_C:
							adj |= DCAF_CONTRAST;
						break;

						case DIKS_SMALL_H:
							adj |= DCAF_HUE;
						break;

						case DIKS_SMALL_S:
							adj |= DCAF_SATURATION;
						break;

						case DIKS_CURSOR_LEFT:
						{
							if (adj)
								adjust_color( video, caps,
										adj, -257 );
							else
								seek( video, caps, -10.0 );
						}
						break;

						case DIKS_CURSOR_RIGHT:
						{
							if (adj)
								adjust_color( video, caps,
										adj, +257 );
							else
								seek( video, caps, +10.0 );
						}
						break;
						
						default:
						break;
					}
				}
				break;

				case DWET_KEYUP:
				{
					switch (tolower( evt.key_symbol ))
					{
						case DIKS_SMALL_B:
							adj ^= DCAF_BRIGHTNESS;
						break;
						
						case DIKS_SMALL_C:
							adj ^= DCAF_CONTRAST;
						break;

						case DIKS_SMALL_H:
							adj ^= DCAF_HUE;
						break;

						case DIKS_SMALL_S:
							adj ^= DCAF_SATURATION;
						break;

						default:
						break;
					}
				}
				break;

				case DWET_BUTTONDOWN:
				{
					if (evt.button != DIBI_LEFT)
						break;

					for (i = 0; i < nwins; i++)
					{
						if (wins[i].id == evt.window_id)
						{
							if (active != wins[i].window)
								moved = 0;
							active = wins[i].window;
						}
					}
					
					if (active)
					{
						layer->GetCursorPosition( layer, 
								&startx, &starty );
						active->GrabPointer( active );
						active->RaiseToTop( active );
					}
				}
				break;

				case DWET_BUTTONUP:
				{
					if (evt.button != DIBI_LEFT)
						break;

					if (active)
					{
						active->UngrabPointer( active );
						active = NULL;
					}
				}
				break;

				case DWET_MOTION:
				{
					if (active)
					{
						moved = 1;
						endx  = evt.cx;
						endy  = evt.cy;
					}
				}
				break;

				case DWET_CLOSE:
				case DWET_DESTROYED:
					goto EXIT;
				break;

				default:
				break;
			}
			
			if (active && moved)
			{
				active->Move( active, endx - startx, endy - starty );
				startx = endx;
				starty = endy;
				moved  = 0;
			}
		}
	}


EXIT:
	if (video)
	{
		video->Stop( video );
		video->Release( video );
	}

	if (buffer)
		buffer->Release( buffer );

	if (wins)
	{
		for (i = 0; i < nwins; i++)
		{
			if (wins[i].surface)
				wins[i].surface->Release( wins[i].surface );
			if (wins[i].window)
				wins[i].window->Release( wins[i].window );
		}
		
		D_FREE( wins );
	}

	if (logo)
		logo->Release( logo );

	if (frame)
		frame->Release( frame );

	if (layer)
		layer->Release( layer );

	if (dfb)
		dfb->Release( dfb );

	return 0;
}

